import numpy as np
from svd_dgesvd import svd_dgesvd
from svd_zgesvd import svd_zgesvd

###############  Some useful fuctions ##########################
def svd(A):
    " SVD of A"
    try:
        X, y, Z = np.linalg.svd(A,full_matrices=0)			
    except np.linalg.LinAlgError:
        if A.dtype == "complex128":
            X, y, Z = svd_zgesvd(A,full_matrices=0)	
        else: 
            X, y, Z = svd_dgesvd(A,full_matrices=0)	
    return X,y,Z

def update(B,s,U,chi_max):
    " Updates the B and s matrices using U_bond and the TEBD protocol "
    d = U.shape[0]
    for i_bond in [0,1]:
        ia = np.mod(i_bond-1,2); ib = np.mod(i_bond,2); ic = np.mod(i_bond+1,2)
        chia = B[ib].shape[1]; chic = B[ic].shape[2]
        
        # Construct theta matrix and time evolution #
        theta = np.tensordot(B[ib],B[ic],axes=(2,1)) # i a j b
        theta = np.tensordot(U,theta,axes=([2,3],[0,2])) # ip jp a b 
        theta = np.tensordot(np.diag(s[ia]),theta,axes=([1,2])) # a ip jp b 
        theta = np.reshape(np.transpose(theta,(1,0,2,3)),(d*chia,d*chic)) # ip a jp b

        # Schmidt deomposition #
        X, Y, Z = svd(theta)			
        chi2 = np.min([np.sum(Y>10.**(-10)), chi_max])	
		
        piv = np.zeros(len(Y), np.bool)
        piv[(np.argsort(Y)[::-1])[:chi2]] = True

        Y = Y[piv]; invsq = np.sqrt(sum(Y**2))
        X = X[:,piv] 
        Z = Z[piv,:]
        
        # Obtain the new values for B and s #
        s[ib] = Y/invsq 
        
        X=np.reshape(X,(d,chia,chi2))
        X = np.transpose(np.tensordot(np.diag(s[ia]**(-1)),X,axes=(1,1)),(1,0,2))
        B[ib] = np.tensordot(X, np.diag(s[ib]),axes=(2,0))        
        B[ic] = np.transpose(np.reshape(Z,(chi2,d,chic)),(1,0,2))

def entanglement_entropy(s):
    " Returns the half chain entanglement entropy "
    S=[]
    for ibond in range(2):
        x=s[ibond][s[ibond]>10**(-20)]**2
        S.append(-np.inner(np.log(x),x))
    return(S)

def bond_expectation_value(B,s,O):
    " Expectation value for a bond operator "
    E=[]
    for ibond in range(2):
        BB = np.tensordot(B[ibond],B[np.mod(ibond+1,2)],axes=(2,1))
        sBB = np.tensordot(np.diag(s[np.mod(ibond-1,2)]),BB,axes=(1,1))
        C = np.tensordot(sBB,O,axes=([1,2],[2,3]))
        sBB=np.conj(sBB)
        E.append(np.squeeze(np.tensordot(sBB,C,axes=([0,3,1,2],[0,1,2,3]))).item()) 
    return E

def site_expectation_value(B,s,O):
    " Expectation value for a site operator "
    E=[]
    for isite in range(0,2):
        sB = np.tensordot(np.diag(s[np.mod(isite-1,2)]),B[isite],axes=(1,1))
        C = np.tensordot(sB,O,axes=(1,0))
        sB=sB.conj()
        E.append(np.squeeze(np.tensordot(sB,C,axes=([0,1,2],[0,2,1]))).item())
    return(E)

